import argparse
import subprocess
from carbontracker.tracker import CarbonTracker
from carbontracker import parser
from carbontracker.report import generate_report_from_log
import ast
import os


def parse_logs(log_dir):
    parser.print_aggregate(log_dir=log_dir)


def generate_report(log_file, output_pdf):
    """Generate a PDF report from a log file"""
    if not os.path.exists(log_file):
        print(f"Error: Log file {log_file} does not exist")
        return
    generate_report_from_log(log_file, output_pdf)
    print(f"Report generated: {output_pdf}")


def main():
    """
    The **carbontracker** CLI allows the user to track the energy consumption and carbon intensity of any program.
    [Make sure that you have relevant permissions before running this.](/#permissions)

    Args:
        --log_dir (path, optional): Log directory. Defaults to `./logs`.
        --api_keys (str, optional): API keys in a dictionary-like format, e.g. `\'{"electricitymaps": "YOUR_KEY"}\'`
        --parse (path, optional): Directory containing the log files to parse.
        --report (path, optional): Generate a PDF report from a log file.
        --output (path, optional): Output path for the generated report. Defaults to 'carbon_report.pdf'
        --sim-cpu (str, optional): Simulated CPU name (overrides detection)
        --sim-cpu-tdp (float, optional): Simulated CPU TDP in Watts
        --sim-cpu-util (float, optional): Simulated CPU utilization (0.0 to 1.0)
        --sim-gpu (str, optional): Simulated GPU name (overrides detection)
        --sim-gpu-watts (float, optional): Simulated GPU power consumption in Watts
        --sim-gpu-util (float, optional): Simulated GPU utilization (0.0 to 1.0)

    Example:
        Tracking the carbon intensity of `script.py`.

            $ carbontracker python script.py

        With example options

            $ carbontracker --log_dir='./logs' --api_keys='{"electricitymaps": "API_KEY_EXAMPLE"}' python script.py

        Using simulated hardware:

            $ carbontracker --sim-cpu "Intel Xeon" --sim-cpu-tdp 150 --sim-gpu "NVIDIA A100" --sim-gpu-watts 400 python script.py

        Parsing logs:

            $ carbontracker --parse ./internal_logs

        Generating a report:

            $ carbontracker --report ./logs/carbontracker.log --output report.pdf
    """

    # Create a parser for the known arguments
    cli_parser = argparse.ArgumentParser(description="CarbonTracker CLI", add_help=True)
    cli_parser.add_argument("--log_dir", type=str, default="./logs", help="Log directory")
    cli_parser.add_argument(
        "--api_keys",
        type=str,
        help="API keys in a dictionary-like format, e.g., "
             '\'{"electricitymaps": "YOUR_KEY"}\'',
        default=None,
    )
    cli_parser.add_argument("--parse", type=str, help="Directory containing the log files to parse.")
    cli_parser.add_argument("--report", type=str, help="Generate a PDF report from a log file.")
    cli_parser.add_argument("--output", type=str, default="carbon_report.pdf", 
                          help="Output path for the generated report.")
    
    # Add simulated hardware arguments
    cli_parser.add_argument("--sim-cpu", type=str, 
                          help="Simulated CPU name (overrides detection). REQUIRED with --sim-cpu-tdp", 
                          default=None)
    cli_parser.add_argument("--sim-cpu-tdp", type=float, 
                          help="Simulated CPU TDP in Watts. REQUIRED when --sim-cpu is specified", 
                          default=None)
    cli_parser.add_argument("--sim-cpu-util", type=float, 
                          help="Simulated CPU utilization (0.0 to 1.0). Defaults to 0.5 if not specified", 
                          default=None)
    cli_parser.add_argument("--sim-gpu", type=str, 
                          help="Simulated GPU name (overrides detection). REQUIRED with --sim-gpu-watts", 
                          default=None)
    cli_parser.add_argument("--sim-gpu-watts", type=float, 
                          help="Simulated GPU power consumption in Watts. REQUIRED when --sim-gpu is specified", 
                          default=None)
    cli_parser.add_argument("--sim-gpu-util", type=float, 
                          help="Simulated GPU utilization (0.0 to 1.0). Defaults to 0.5 if not specified", 
                          default=None)

    # Parse known arguments only
    known_args, remaining_args = cli_parser.parse_known_args()

    # Check if the --parse argument is provided
    if known_args.parse:
        parse_logs(known_args.parse)
        return

    # Check if the --report argument is provided
    if known_args.report:
        generate_report(known_args.report, known_args.output)
        return

    # Parse the API keys string into a dictionary
    api_keys = ast.literal_eval(known_args.api_keys) if known_args.api_keys else None

    tracker = CarbonTracker(
        epochs=1, 
        log_dir=known_args.log_dir, 
        epochs_before_pred=0, 
        api_keys=api_keys,
        sim_cpu=known_args.sim_cpu,
        sim_cpu_tdp=known_args.sim_cpu_tdp,
        sim_cpu_util=known_args.sim_cpu_util,
        sim_gpu=known_args.sim_gpu,
        sim_gpu_watts=known_args.sim_gpu_watts,
        sim_gpu_util=known_args.sim_gpu_util
    )
    tracker.epoch_start()

    # The remaining_args are considered as the command to execute
    if remaining_args:
        try:
            # Execute the command
            subprocess.run(remaining_args, check=True)
        except subprocess.CalledProcessError as e:
            print(f"Error executing command: {' '.join(remaining_args)}")
            print(f"Subprocess error: {e}")

    tracker.epoch_end()
    tracker.stop()


if __name__ == "__main__":
    main()


import os
import re

import numpy as np

from carbontracker import exceptions
from typing import Dict, Union, List


def parse_all_logs(log_dir):
    """
    Parse all logs in directory.

    Args:
        log_dir (str): Directory of logs

    Returns:
        (dict[]): List of log entries of shape

                {
                    "output_filename": str,
                    "standard_filename": str,
                    "components": dict, # See parse_logs
                    "early_stop": bool,
                    "actual": dict | None, # See get_consumption
                    "pred": dict | None, # See get_consumption
                }
    """
    logs = []
    output_logs, std_logs = get_all_logs(log_dir)

    for out, std in zip(output_logs, std_logs):
        with open(std, "r") as f:
            std_log_data = f.read()

        with open(out, "r") as f:
            output_log_data = f.read()

        actual, pred = get_consumption(output_log_data)
        early_stop = get_early_stop(std_log_data)
        entry = {
            "output_filename": out,
            "standard_filename": std,
            "components": parse_logs(log_dir, std, out),
            "early_stop": early_stop,
            "actual": actual,
            "pred": pred,
        }
        logs.append(entry)

    return logs


def parse_logs(log_dir, std_log_file=None, output_log_file=None):
    """
    Parse logs in log_dir (defaults to most recent logs).

    Args:
        log_dir (str): Directory of logs
        std_log_file (str, optional): Log file to read. Defaults to most recent logs.
        output_log_file (str, optional): Deprecated

    Returns:
        (dict): Dictionary of shape

                {
                        [component name]: {
                            "avg_power_usages (W)": NDArray | None,
                            "avg_energy_usages (J)": NDArray | None,
                            "epoch_durations (s)": NDArray | None,
                            "devices": str[],
                        }
                }

            where `[component name]` is either `"gpu"` or `"cpu"`.
            Return value can contain both `"gpu"` and `"cpu"` field.
    """
    if std_log_file is None or output_log_file is None:
        std_log_file, output_log_file = get_most_recent_logs(log_dir)

    with open(std_log_file, "r") as f:
        std_log_data = f.read()

    epoch_durations = get_epoch_durations(std_log_data)
    avg_power_usages = get_avg_power_usages(std_log_data)
    devices = get_devices(std_log_data)

    components = {}
    for comp, devices in devices.items():
        power_usages = (
            np.array(avg_power_usages[comp]) if len(avg_power_usages) != 0 else None
        )
        durations = np.array(epoch_durations) if len(epoch_durations) != 0 else None
        if power_usages is None or durations is None:
            energy_usages = None
        else:
            if power_usages.size != durations.size:
                raise exceptions.MismatchedEpochsError(
                    f"Found {power_usages.size} power measurements and {durations.size} duration measurements. "
                    "Expected equal number of measurements."
                )
            energy_usages = (power_usages.T * durations).T
        measurements = {
            "avg_power_usages (W)": power_usages,
            "avg_energy_usages (J)": energy_usages,
            "epoch_durations (s)": durations,
            "devices": devices,
        }
        components[comp] = measurements

    return components


def get_consumption(output_log_data: str):
    """
    Gets actual and predicted energy consumption, CO2eq and equivalence statements from output_log_data using regular expressions.

    Args:
        output_log_data (str): Log data to search through.

    Returns:
        actual (dict | None): Actual consumption

        pred (dict | None): Predicted consumption

            Both `actual` and `pred` has the shape:

                {
                    "epochs": int,
                    "duration (s)": int,
                    "energy (kWh)": float | None,
                    "co2eq (g)": float | None,
                    "equivalents": equivalents,
                }
    """
    actual_re = re.compile(
        r"(?i)Actual consumption"
        r"(?:\s*for\s+\d+\s+epochs)?"
        r"[\s\S]*?Time:\s*(.*)\n\s*Energy:\s*(.*)\s+kWh"
        r"[\s\S]*?CO2eq:\s*(.*)\s+g"
        r"(?:\s*This is equivalent to:\s*([\s\S]*?))?(?=\d{4}-\d{2}-\d{2}|\Z)"
    )
    pred_re = re.compile(
        r"(?i)Predicted consumption for (\d*) epoch\(s\):"
        r"[\s\S]*?Time:\s*(.*)\n\s*Energy:\s*(.*)\s+kWh"
        r"[\s\S]*?CO2eq:\s*(.*)\s+g"
        r"(?:\s*This is equivalent to:\s*([\s\S]*?))?(?=\d{4}-\d{2}-\d{2}|\Z)"
    )
    actual_match = re.search(actual_re, output_log_data)
    pred_match = re.search(pred_re, output_log_data)
    actual = extract_measurements(actual_match)
    pred = extract_measurements(pred_match)
    return actual, pred


def get_early_stop(std_log_data: str) -> bool:
    early_stop_re = re.compile(r"(?i)Training was interrupted")
    early_stop = re.findall(early_stop_re, std_log_data)
    return bool(early_stop)

def extract_measurements(match):
    if not match:
        return None
    match = match.groups()
    if len(match) == 4:
        match = [1] + list(match)
    epochs = int(match[0])
    duration = get_time(match[1])
    energy, co2eq, equivalents = get_stats(match)
    measurements = {
        "epochs": epochs,
        "duration (s)": duration,
        "energy (kWh)": energy,
        "co2eq (g)": co2eq,
        "equivalents": equivalents,
    }
    return measurements


def get_time(time_str: str) -> Union[float, None]:
    duration_re = re.compile(r"(\d+):(\d{2}):(\d\d?(?:.\d{2})?)")
    match = re.search(duration_re, time_str)
    if not match:
        return None
    match = match.groups()
    duration = float(match[0]) * 60 * 60 + float(match[1]) * 60 + float(match[2])
    return duration


def print_aggregate(log_dir):
    """
    Prints the aggregate consumption in all log files in log_dir to stdout. See `get_aggregate`.

    Args:
        log_dir (str): Directory of logs
    """
    energy, co2eq, equivalents = aggregate_consumption(log_dir)

    equivalents_p = " or ".join([f"{v:.16f} {k}" for k, v in equivalents.items()])

    printable = f"The training of models in this work is estimated to use {energy:.16f} kWh of electricity contributing to {co2eq / 1000:.16f} kg of CO2eq. "
    if equivalents_p:
        printable += f"This is equivalent to {equivalents_p}. "

    printable += "Measured by carbontracker (https://github.com/lfwa/carbontracker)."

    print(printable)


def aggregate_consumption(log_dir):
    """
    Aggregate consumption in all log files in specified log_dir.

    Args:
        log_dir (str): Directory of logs

    Returns:
        total_energy (float): Total energy (kWh) of all logs
        total_co2 (float): Total CO2eq (gCO2eq) of all logs
        total_equivalents (float): Total energy of all logs
    """
    output_logs, std_logs = get_all_logs(log_dir=log_dir)

    total_energy = 0
    total_co2eq = 0
    total_equivalents = {}

    for output_log, std_log in zip(output_logs, std_logs):
        with open(output_log, "r") as f:
            output_data = f.read()
        with open(std_log, "r") as f:
            std_data = f.read()

        actual, pred = get_consumption(output_data)
        early_stop = get_early_stop(std_data)

        if actual is None and pred is None:
            continue
        elif actual is None and pred is not None:
            energy = pred["energy (kWh)"]
            co2eq = pred["co2eq (g)"]
            equivalents = pred["equivalents"]
        elif pred is None and actual is not None:
            energy = actual["energy (kWh)"]
            co2eq = actual["co2eq (g)"]
            equivalents = actual["equivalents"]
        # Both actual and pred is available
        elif pred is not None and actual is not None:
            actual_epochs = actual["epochs"]
            pred_epochs = pred["epochs"]
            if early_stop or actual_epochs == pred_epochs:
                energy = actual["energy (kWh)"]
                co2eq = actual["co2eq (g)"]
                equivalents = actual["equivalents"]
            else:
                energy = pred["energy (kWh)"]
                co2eq = pred["co2eq (g)"]
                equivalents = pred["equivalents"]
        else:
            continue  # unreachable case

        total_energy += energy
        if not np.isnan(co2eq):
            total_co2eq += co2eq
        if equivalents is not None:
            for key, value in equivalents.items():
                total_equivalents[key] = total_equivalents.get(key, 0) + value

    return total_energy, total_co2eq, total_equivalents


def get_stats(groups):
    energy = float(groups[2])
    co2eq = float(groups[3])
    if len(groups) >= 5 and groups[4] is not None:
        equivalents = parse_equivalents(groups[4])
    else:
        equivalents = None
    return energy, co2eq, equivalents


def parse_equivalents(lines):
    equivalents = {}
    lines = lines.split("\n")
    for line in lines:
        tup = line.split(" ", 1)
        if tup[0] and tup[1]:
            try:
                equivalents[tup[1].strip()] = float(tup[0].strip())
            except ValueError as e:
                print(
                    f"Warning: Unable to convert '{tup[0]}' to float. Skipping this equivalent."
                )
                continue
    return equivalents


def get_all_logs(log_dir):
    """
    Get all output and standard logs in log_dir.

    Args:
        log_dir (str): Directory of logs

    Returns:
        std_logs (list[str]): List of file names of standard logs
        output_logs (list[str]): List of file names of output logs

    Raises:
        MismatchedLogFilesError: Thrown if there exists standard log files that cannot be matched with an output log file or vice versa.
    """
    files = [
        os.path.join(log_dir, f)
        for f in os.listdir(log_dir)
        if os.path.isfile(os.path.join(log_dir, f))
        and os.path.getsize(os.path.join(log_dir, f)) > 0
    ]
    output_re = re.compile(r".*carbontracker_output.log")
    std_re = re.compile(r".*carbontracker.log")
    output_logs = sorted(list(filter(output_re.match, files)))
    std_logs = sorted(list(filter(std_re.match, files)))
    if len(output_logs) != len(std_logs):
        # Try to remove the files with no matching output/std logs
        op_fn = [f.split("_carbontracker")[0] for f in output_logs]
        std_fn = [f.split("_carbontracker")[0] for f in std_logs]
        if len(std_logs) > len(output_logs):
            missing_logs = list(set(std_fn) - set(op_fn))
            [std_logs.remove(f + "_carbontracker.log") for f in missing_logs]
        else:
            missing_logs = list(set(op_fn) - set(std_fn))
            [output_logs.remove(f + "_carbontracker_output.log") for f in missing_logs]
        ### Even after removal if then there is a mismatch, then throw the error
        if len(output_logs) != len(std_logs):
            raise exceptions.MismatchedLogFilesError(
                f"Found {len(output_logs)} output logs and {len(std_logs)} "
                "standard logs. Expected equal number of logs."
            )
    return output_logs, std_logs


def get_devices(std_log_data: str) -> Dict[str, List[str]]:
    """
    Retrieve dictionary of components with their device(s).

    Args:
        std_log_data (str): Log data to parse

    Returns:
        (dict): Dictionary with devices per component of shape

                {
                        [component]: ["device1", "device2"]
                }

            Where `[component]` is the component name and `"device1"`, `"device2"` are device names.
    """
    comp_re = re.compile(r"The following components were found:(.*)\n")
    device_re = re.compile(r" (.*?) with device\(s\) (.*?)\.")
    # Take first match as we only expect one.
    match = re.findall(comp_re, std_log_data)
    if not match:
        return {}
    device_matches = re.findall(device_re, match[0])
    devices = {}

    for comp, device_str in device_matches:
        dev = device_str.split(",")
        devices[comp.lower()] = dev

    return devices


def get_epoch_durations(std_log_data):
    """
    Retrieve epoch durations (s).

    Args:
        std_log_data (str): Log to parse

    Returns:
        (list[float]): List of epoch durations (s)
    """
    duration_re = re.compile(r"Duration: (\d+):(\d{2}):(\d\d?(?:.\d{2})?)")
    matches = re.findall(duration_re, std_log_data)
    epoch_durations = [
        float(h) * 60 * 60 + float(m) * 60 + float(s) for h, m, s in matches
    ]
    return epoch_durations


def get_avg_power_usages(std_log_data):
    """
    Retrieve average power usages for each epoch (W).

    Args:
        std_log_data (str): Log to parse

    Returns:
        (dict): Dictionary containing list of average power usages for each epoch per component. Has shape:
                {
                        [component name]: list[list[float]]
                }
    """
    power_re = re.compile(r"Average power usage \(W\) for (.+): (\[?[0-9\.]+\]?|None)")
    matches = re.findall(power_re, std_log_data)
    components = list(set([comp for comp, _ in matches]))
    avg_power_usages = {}

    for component in components:
        powers: list[list[float]] = []
        for comp, power in matches:
            if comp == component:
                if power == "None":
                    powers.append([0.0])
                    continue
                else:
                    p_list = power.strip("[").strip("]").split(" ")
                    p_power = [float(num) for num in p_list if num != ""]
                    powers.append(p_power)
        avg_power_usages[component] = powers

    return avg_power_usages


def get_most_recent_logs(log_dir):
    """
    Retrieve the file names of the most recent standard and output logs.

    Args:
        log_dir (str): Directory of logs

    Returns:
        std_log (str): File name of latest standard log
        output_log (str): File name of latest output log
    """
    # Get all files in log_dir.
    files = [
        os.path.join(log_dir, f)
        for f in os.listdir(log_dir)
        if os.path.isfile(os.path.join(log_dir, f))
    ]
    # Find output and standard logs and sort by modified date.
    output_re = re.compile(r".*carbontracker_output.log")
    std_re = re.compile(r".*carbontracker.log")
    output_logs = list(filter(output_re.match, files))
    std_logs = list(filter(std_re.match, files))
    output_logs.sort(key=os.path.getmtime)
    std_logs.sort(key=os.path.getmtime)

    return std_logs[-1], output_logs[-1]



import os
import sys
import time
import traceback
import psutil
import math
from threading import Thread, Event
from typing import List, Union

import numpy as np
from random import randint

from carbontracker import constants
from carbontracker import loggerutil
from carbontracker import predictor
from carbontracker import exceptions
from carbontracker.components import component
from carbontracker.components.component import Component
from carbontracker.emissions.intensity import intensity
from carbontracker.emissions.conversion import co2eq
from carbontracker.emissions.intensity.fetchers import electricitymaps


class CarbonIntensityThread(Thread):
    """Sleeper thread to update Carbon Intensity every 15 minutes."""

    def __init__(self, logger, stop_event, update_interval: Union[float, int] = 900):
        super(CarbonIntensityThread, self).__init__()
        self.name = "CarbonIntensityThread"
        self.logger = logger
        self.update_interval: Union[float, int] = update_interval
        self.daemon = True
        self.stop_event = stop_event
        self.carbon_intensities = []

        self.start()

    def run(self):
        try:
            self._fetch_carbon_intensity()
            while not self.stop_event.wait(self.update_interval):
                self._fetch_carbon_intensity()
        except Exception:
            err_str = traceback.format_exc()
            self.logger.err_warn(err_str)

    def _fetch_carbon_intensity(self):
        ci = intensity.carbon_intensity(self.logger)
        if (
            ci.success
            and isinstance(ci.carbon_intensity, (int, float))
            and not np.isnan(ci.carbon_intensity)
        ):
            self.carbon_intensities.append(ci)
            self.logger.info(f"Carbon intensity: {ci.carbon_intensity:.2f} gCO2/kWh at {ci.address}")

    def predict_carbon_intensity(self, pred_time_dur):
        ci = intensity.carbon_intensity(self.logger, time_dur=pred_time_dur)
        weighted_intensities = [
            ci.carbon_intensity for ci in self.carbon_intensities
        ] + [ci.carbon_intensity]

        # Account for measured intensities by taking weighted average.
        weight = math.floor(pred_time_dur / self.update_interval)

        for _ in range(weight):
            weighted_intensities.append(ci.carbon_intensity)

        ci.carbon_intensity = np.mean(weighted_intensities)
        intensity.set_carbon_intensity_message(ci, pred_time_dur)

        self.logger.info(ci.message)
        self.logger.output(ci.message, verbose_level=2)

        return ci

    def average_carbon_intensity(self):
        if not self.carbon_intensities:
            ci = intensity.carbon_intensity(self.logger)
            self.carbon_intensities.append(ci)

        # Ensure that we have some carbon intensities.
        assert self.carbon_intensities

        location = self.carbon_intensities[-1].address
        intensities = [ci.carbon_intensity for ci in self.carbon_intensities]
        avg_intensity = np.mean(intensities)
        msg = (
            f"Average carbon intensity during training was {avg_intensity:.2f}"
            f" gCO2/kWh at detected location: {location}."
        )
        avg_ci = intensity.CarbonIntensity(
            carbon_intensity=avg_intensity, message=msg, success=True
        )

        self.logger.info(
            "Carbon intensities (gCO2/kWh) fetched every "
            f"{self.update_interval} s at detected location {location}: "
            f"{intensities}"
        )
        self.logger.info(avg_ci.message)
        self.logger.output(avg_ci.message, verbose_level=2)

        return avg_ci


class CarbonTrackerThread(Thread):
    """Thread to fetch consumptions"""

    def __init__(
        self,
        components: List[Component],
        logger,
        ignore_errors,
        delete,
        update_interval: Union[int, float] = 1,
    ):
        super(CarbonTrackerThread, self).__init__()
        self.cur_epoch_time = time.time()
        self.name = "CarbonTrackerThread"
        self.delete = delete
        self.components = components
        self.update_interval = update_interval
        self.ignore_errors = ignore_errors
        self.logger = logger
        self.epoch_times = []
        self.running = True
        self.measuring_event = Event()
        self.epoch_counter = 0
        self.daemon = True

        self.start()

    def run(self):
        """Thread's activity."""
        try:
            self.begin()
            while self.running:
                # Wait for the measuring_event to be set
                self.measuring_event.wait()
                self._collect_measurements()
                time.sleep(self.update_interval)

            # Shutdown in thread's activity instead of epoch_end() to ensure
            # that we only shutdown after last measurement.
            self._components_shutdown()
        except Exception as e:
            self._handle_error(e)

    def begin(self):
        self._components_remove_unavailable()
        self._components_init()
        self._log_components_info()
        self.logger.info("Monitoring thread started.")

    def stop(self):
        if not self.running:
            return

        self.running = False
        self.logger.info("Monitoring thread ended.")
        self.logger.output("Finished monitoring.", verbose_level=1)

    def epoch_start(self):
        self.epoch_counter += 1
        self.cur_epoch_time = time.time()
        self.measuring_event.set()  # Set the event to start measuring

    def epoch_end(self):
        self.measuring_event.clear()  # Clear the event to stop measuring
        self.epoch_times.append(time.time() - self.cur_epoch_time)
        self._log_epoch_measurements()

    def _log_components_info(self):
        log = ["The following components were found:"]
        for comp in self.components:
            name = comp.name.upper()
            devices = ", ".join(comp.devices())
            log.append(f"{name} with device(s) {devices}.")
        log_str = " ".join(log)
        self.logger.info(log_str)
        self.logger.output(log_str, verbose_level=1)

    def _log_epoch_measurements(self):
        self.logger.info(f"Epoch {self.epoch_counter}:")
        duration = self.epoch_times[-1]
        self.logger.info(
            f"Duration: {loggerutil.convert_to_timestring(duration, True)}"
        )
        for comp in self.components:
            if comp.power_usages and comp.power_usages[-1]:
                power_avg = np.mean(comp.power_usages[-1], axis=0)
                # If np.mean is calculated during a measurement, it will get an
                # empty list and return nan, if this is the case we take the
                #  previous measurement.
                # TODO: Use semaphores to wait for measurement to finish.
                if np.isnan(power_avg).all():
                    power_avg = (
                        np.mean(comp.power_usages[-2], axis=0)
                        if len(comp.power_usages) >= 2
                        else None
                    )
            else:
                self.logger.err_warn(
                    "Epoch duration is too short for a measurement to be " "collected."
                )
                power_avg = None

            self.logger.info(f"Average power usage (W) for {comp.name}: {power_avg}")

    def _components_remove_unavailable(self):
        self.components = [cmp for cmp in self.components if cmp.available()]
        if not self.components:
            raise exceptions.NoComponentsAvailableError()

    def _components_init(self):
        for comp in self.components:
            comp.init()

    def _components_shutdown(self):
        for comp in self.components:
            comp.shutdown()

    def _collect_measurements(self):
        """Collect one round of measurements."""
        for comp in self.components:
            comp.collect_power_usage(self.epoch_counter)

    def total_energy_per_epoch(self):
        """Retrieves total energy (kWh) per epoch used by all components
        including PUE."""
        total_energy = np.zeros(len(self.epoch_times))
        for comp in self.components:
            energy_usage = comp.energy_usage(self.epoch_times)
            total_energy += energy_usage
        return total_energy * constants.PUE_2023

    def _handle_error(self, error):
        err_str = traceback.format_exc()
        if self.ignore_errors:
            err_str = (
                f"Ignored error: {err_str}Continued training without " "monitoring..."
            )

        self.logger.err_critical(err_str)
        self.logger.output(err_str)

        if self.ignore_errors:
            # Stop monitoring but continue training.
            self.delete()
        else:
            os._exit(70)


class CarbonTracker:
    """

    The CarbonTracker class is the main interface for starting, stopping and reporting through **carbontracker**.

    Args:
        epochs (int): Total epochs of your training loop.
        api_keys (dict, optional): Dictionary of Carbon Intensity API keys following the {name:key} format. Can also be set using `CarbonTracker.set_api_keys`

            Example: `{ \\"electricitymaps\\": \\"abcdefg\\" }`
        epochs_before_pred (int, optional): Epochs to monitor before outputting predicted consumption. Set to -1 for all epochs. Set to 0 for no prediction.
        monitor_epochs (int, optional): Total number of epochs to monitor. Outputs actual consumption when reached. Set to -1 for all epochs. Cannot be less than `epochs_before_pred` or equal to 0.
        update_interval (int, optional): Interval in seconds between power usage measurements are taken by sleeper thread.
        interpretable (bool, optional): If set to `True` then the CO2eq are also converted to interpretable numbers such as the equivalent distance travelled in a car, etc. Otherwise, no conversions are done.
        stop_and_confirm (bool, optional): If set to `True` then the main thread (with your training loop) is paused after epochs_before_pred epochs to output the prediction and the user will need to confirm to continue training. Otherwise, prediction is output and training is continued instantly.
        ignore_errors (bool, optional): If set to `True` then all errors will cause energy monitoring to be stopped and training will continue. Otherwise, training will be interrupted as with regular errors.
        components (str, optional): Comma-separated string of which components to monitor. Options are: `"all"`, `"gpu"`, `"cpu"`, or `"gpu,cpu"`.
        devices_by_pid (bool, optional): If `True`, only devices (under the chosen components) running processes associated with the main process are measured. If False, all available devices are measured. Note that this requires your devices to have active processes before instantiating the CarbonTracker class.
        log_dir (str, optional): Path to the desired directory to write log files. If `None`, then no logging will be done.
        log_file_prefix (str, optional): Prefix to add to the log file name.
        verbose (int, optional): Sets the level of verbosity.
        decimal_precision (int, optional): Desired decimal precision of reported values.
        sim_cpu (float, optional): Custom CPU value for components.
        sim_cpu_tdp (float, optional): Custom TDP value for components.
        sim_cpu_util (float, optional): Custom CPU utilization for components.
        sim_gpu (float, optional): Custom GPU value for components.
        sim_gpu_watts (float, optional): Custom GPU Watts value for components.
        sim_gpu_util (float, optional): Custom GPU utilization for components.

    Example:
        Tracking the carbon intensity of PyTorch model training:

            from carbontracker.tracker import CarbonTracker

            tracker = CarbonTracker(epochs=max_epochs)
            # Training loop.
            for epoch in range(max_epochs):
                tracker.epoch_start()
                # Your model training.
                tracker.epoch_end()

            # Optional: Add a stop in case of early termination before all monitor_epochs has
            # been monitored to ensure that actual consumption is reported.
            tracker.stop()

    """

    def __init__(
        self,
        epochs,
        epochs_before_pred=1,
        monitor_epochs=-1,
        update_interval=1,
        interpretable=True,
        stop_and_confirm=False,
        ignore_errors=False,
        components="all",
        devices_by_pid=False,
        log_dir=None,
        log_file_prefix="",
        verbose=1,
        decimal_precision=12,
        api_keys=None,
        sim_cpu=None,
        sim_cpu_tdp=None,
        sim_cpu_util=None,
        sim_gpu=None,
        sim_gpu_watts=None,
        sim_gpu_util=None
    ):
        """Initialize CarbonTracker.

        Args:
            epochs (int): Number of epochs to monitor.
            epochs_before_pred (int, optional): Number of epochs to monitor before making predictions. Defaults to 1.
            monitor_epochs (int, optional): Number of epochs to monitor. Defaults to -1.
            update_interval (int, optional): Interval in seconds between measurements. Defaults to 1.
            interpretable (bool, optional): Whether to make predictions interpretable. Defaults to True.
            stop_and_confirm (bool, optional): Whether to stop and confirm before making predictions. Defaults to False.
            ignore_errors (bool, optional): Whether to ignore errors. Defaults to False.
            components (str, optional): Components to monitor. Defaults to "all".
            devices_by_pid (bool, optional): Whether to monitor devices by PID. Defaults to False.
            log_dir (str, optional): Directory to store logs. Defaults to None.
            log_file_prefix (str, optional): Prefix for log files. Defaults to "".
            verbose (int, optional): Verbosity level. Defaults to 1.
            decimal_precision (int, optional): Decimal precision for measurements. Defaults to 12.
            api_keys (dict, optional): API keys for external services. Defaults to None.
            sim_cpu (str, optional): Simulated CPU name. Defaults to None.
            sim_cpu_tdp (float, optional): Simulated CPU TDP in Watts. Defaults to None.
            sim_cpu_util (float, optional): Simulated CPU utilization. Defaults to None.
            sim_gpu (str, optional): Simulated GPU name. Defaults to None.
            sim_gpu_watts (float, optional): Simulated GPU power consumption in Watts. Defaults to None.
            sim_gpu_util (float, optional): Simulated GPU utilization. Defaults to None.
        """
        # Add validation for monitor_epochs
        if monitor_epochs != -1:
            if monitor_epochs < epochs_before_pred:
                raise ValueError("monitor_epochs cannot be less than epochs_before_pred")
            if monitor_epochs == 0:
                raise ValueError("monitor_epochs cannot be zero")

        # Validate simulated component configurations
        if sim_cpu is not None and sim_cpu_tdp is None:
            raise ValueError("When using simulated CPU (sim_cpu), you must also specify the CPU TDP (sim_cpu_tdp)")
        if sim_gpu is not None and sim_gpu_watts is None:
            raise ValueError("When using simulated GPU (sim_gpu), you must also specify the GPU power consumption (sim_gpu_watts)")

        self.epochs = epochs
        self.epochs_before_pred = epochs_before_pred
        self.monitor_epochs = monitor_epochs
        self.update_interval = update_interval
        self.interpretable = interpretable
        self.stop_and_confirm = stop_and_confirm
        self.ignore_errors = ignore_errors
        self.components = components
        self.devices_by_pid = devices_by_pid
        self.log_dir = log_dir
        self.log_file_prefix = log_file_prefix
        self.verbose = verbose
        self.decimal_precision = decimal_precision
        self.api_keys = api_keys
        self.sim_cpu = sim_cpu
        self.sim_cpu_tdp = sim_cpu_tdp
        self.sim_cpu_util = sim_cpu_util
        self.sim_gpu = sim_gpu
        self.sim_gpu_watts = sim_gpu_watts
        self.sim_gpu_util = sim_gpu_util
        self.deleted = False
        self.epoch_counter = 0

        try:
            pids = self._get_pids()
            self.logger = loggerutil.Logger(
                log_dir=log_dir,
                verbose=verbose,
                log_prefix=log_file_prefix,
                logger_id=str(randint(1, 999999)),
            )
            self.tracker = CarbonTrackerThread(
                delete=self._delete,
                components=component.create_components(
                    components=components, pids=pids, devices_by_pid=devices_by_pid, logger=self.logger, sim_cpu=self.sim_cpu, sim_cpu_tdp=self.sim_cpu_tdp, sim_cpu_util=self.sim_cpu_util, sim_gpu=self.sim_gpu, sim_gpu_watts=self.sim_gpu_watts, sim_gpu_util=self.sim_gpu_util
                ),
                logger=self.logger,
                ignore_errors=ignore_errors,
                update_interval=update_interval,
            )
            self.intensity_stopper = Event()
            self.intensity_updater = CarbonIntensityThread(
                self.logger, self.intensity_stopper
            )
        except Exception as e:
            self._handle_error(e)

    def epoch_start(self):
        """
        Starts tracking energy consumption for current epoch. Call in the beginning of training loop.
        """
        if self.deleted:
            return

        try:
            self.tracker.epoch_start()
            self.epoch_counter += 1
        except Exception as e:
            self._handle_error(e)

    def epoch_end(self):
        """
        Ends tracking energy consumption for current epoch. Call in the end of training loop.
        """
        if self.deleted:
            return

        try:
            self.tracker.epoch_end()

            if self.epoch_counter == self.epochs_before_pred:
                self._output_pred()
                if self.stop_and_confirm:
                    self._user_query()

            if self.epoch_counter == self.monitor_epochs:
                self._output_actual()

            if self.epoch_counter == self.monitor_epochs:
                self._delete()
        except Exception as e:
            self._handle_error(e)

    def stop(self):
        """Ensure that tracker is stopped and deleted. E.g. use with early
        stopping, where not all monitor_epochs have been run."""
        if self.deleted:
            return
        self.logger.info(
            f"Training was interrupted before all {self.monitor_epochs} epochs"
            " were monitored."
        )
        # Decrement epoch_counter with 1 since measurements for ultimate epoch
        # was interrupted and is not accounted for.
        self.epoch_counter -= 1
        self._output_actual()
        self._delete()

    def set_api_keys(self, api_dict):
        """Set API keys (given as {name:key}) for carbon intensity fetchers."""
        try:
            for name, key in api_dict.items():
                if name.lower() == "electricitymaps":
                    electricitymaps.ElectricityMap.set_api_key(key)
                else:
                    raise exceptions.FetcherNameError(
                        f"Invalid API name '{name}' given."
                    )
        except Exception as e:
            self._handle_error(e)

    def _handle_error(self, error):
        err_str = traceback.format_exc()
        if self.ignore_errors:
            err_str = (
                f"Ignored error: {err_str}Continued training without " "monitoring..."
            )

        self.logger.err_critical(err_str)
        self.logger.output(err_str)

        if self.ignore_errors:
            # Stop monitoring but continue training.
            self._delete()
        else:
            sys.exit(70)

    def _output_energy(self, description, time, energy, co2eq, conversions):
        precision = self.decimal_precision
        output = (
            f"\n{description}\n"
            f"\tTime:\t{loggerutil.convert_to_timestring(time)}\n"
            f"\tEnergy:\t{energy:.{precision}f} kWh\n"
            f"\tCO2eq:\t{co2eq:.{precision}f} g"
        )

        if conversions:
            conv_str = "\n\tThis is equivalent to:"
            for units, unit in conversions:
                conv_str += f"\n\t{units:.{precision}f} {unit}"
            output += conv_str

        self.logger.output(output, verbose_level=1)

    def _output_actual(self):
        """Output actual usage so far."""
        energy_usages = self.tracker.total_energy_per_epoch()
        energy = energy_usages.sum()
        times = self.tracker.epoch_times
        time = np.sum(times)
        _co2eq = self._co2eq(energy)
        conversions = co2eq.convert(_co2eq) if self.interpretable else None
        if self.epochs_before_pred == 0:
            self._output_energy(
                "Actual consumption:", time, energy, _co2eq, conversions
            )
        else:
            self._output_energy(
                f"Actual consumption for {self.epoch_counter} epoch(s):",
                time,
                energy,
                _co2eq,
                conversions,
            )

    def _output_pred(self):
        """Output predicted usage for full training epochs."""
        epoch_energy_usages = self.tracker.total_energy_per_epoch()
        epoch_times = self.tracker.epoch_times
        pred_energy = predictor.predict_energy(self.epochs, epoch_energy_usages)
        pred_time = predictor.predict_time(self.epochs, epoch_times)
        pred_co2eq = self._co2eq(pred_energy, pred_time)
        conversions = co2eq.convert(pred_co2eq) if self.interpretable else None

        self._output_energy(
            f"Predicted consumption for {self.epochs} epoch(s):",
            pred_time,
            pred_energy,
            pred_co2eq,
            conversions,
        )

    def _co2eq(self, energy_usage, pred_time_dur=None):
        """ "Returns the CO2eq (g) of the energy usage (kWh)."""
        if pred_time_dur:
            ci = self.intensity_updater.predict_carbon_intensity(pred_time_dur)
        else:
            ci = self.intensity_updater.average_carbon_intensity()
        co2eq = energy_usage * ci.carbon_intensity
        return co2eq

    def _user_query(self):
        self.logger.output("Continue training (y/n)?")
        user_input = input().lower()
        self._check_input(user_input)

    def _check_input(self, user_input: str):
        if user_input == "y":
            self.logger.output("Continuing...")
            return
        elif user_input == "n":
            self.logger.info("Session ended by user.")
            self.logger.output("Quitting...")
            sys.exit(0)
        else:
            self.logger.output("Input not recognized. Try again (y/n):")
            user_input = input().lower()
            self._check_input(user_input)

    def _delete(self):
        self.tracker.stop()
        self.intensity_stopper.set()
        del self.logger
        del self.tracker
        del self.intensity_updater
        del self.intensity_stopper
        self.deleted = True

    def _get_pids(self) -> List[int]:
        """Get current process id and all children process ids."""
        process = psutil.Process()
        pids = [process.pid] + [child.pid for child in process.children(recursive=True)]
        return pids



import re
from datetime import datetime
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from reportlab.lib import colors
from reportlab.lib.pagesizes import letter
from reportlab.platypus import SimpleDocTemplate, Paragraph, Spacer, Table, TableStyle, Image
from reportlab.lib.styles import getSampleStyleSheet, ParagraphStyle
from reportlab.lib.units import inch
import io

# Set the style for all plots
sns.set_style("whitegrid")
sns.set_palette("husl")
plt.rcParams.update({
    'figure.dpi': 600,
    'savefig.dpi': 600,
    'font.size': 14,
    'axes.labelsize': 14,
    'axes.titlesize': 14,
    'xtick.labelsize': 9,
    'ytick.labelsize': 9,
    'legend.fontsize': 9,
    'figure.titlesize': 13
})

def format_duration(seconds):
    """Convert seconds to a human-readable duration string."""
    hours = int(seconds // 3600)
    minutes = int((seconds % 3600) // 60)
    seconds = int(seconds % 60)
    
    parts = []
    if hours > 0:
        parts.append(f"{hours}h")
    if minutes > 0:
        parts.append(f"{minutes}min")
    if seconds > 0 or not parts:
        parts.append(f"{seconds}s")
    
    return " ".join(parts)

class LogParser:
    def __init__(self, log_content):
        self.log_content = log_content
        self.epochs = []
        self.carbon_intensity = None
        self.location = None
        self.version = None
        self.pue = None
        self.components = None
        self.start_time = None
        self.end_time = None
        self._parse_log()
    
    def _parse_log(self):
        # Parse version and PUE info
        version_match = re.search(r'carbontracker version ([\d\.]+)', self.log_content)
        if version_match:
            self.version = version_match.group(1)
            
        pue_match = re.search(r'PUE coefficient of ([\d\.]+)', self.log_content)
        if pue_match:
            self.pue = float(pue_match.group(1))
        
        # Parse components
        components_match = re.search(r'The following components were found: (.*)', self.log_content)
        if components_match:
            self.components = components_match.group(1)
        
        # Parse epochs
        epoch_pattern = r'Epoch (\d+):\n.*?Duration: ([\d:\.]+)\n.*?Average power usage \(W\) for gpu: ([\d\.]+)\n.*?Average power usage \(W\) for cpu: ([\d\.]+)'
        epoch_matches = re.finditer(epoch_pattern, self.log_content, re.DOTALL)
        
        for match in epoch_matches:
            epoch_num = int(match.group(1))
            duration = self._parse_duration(match.group(2))
            gpu_power = float(match.group(3))
            cpu_power = float(match.group(4))
            
            self.epochs.append({
                'epoch': epoch_num,
                'duration': duration,
                'gpu_power': gpu_power,
                'cpu_power': cpu_power,
                'total_power': gpu_power + cpu_power
            })
        
        # Parse carbon intensity
        ci_match = re.search(r'Average carbon intensity during training was ([\d\.]+) gCO2/kWh at detected location: (.*)', self.log_content)
        if ci_match:
            self.carbon_intensity = float(ci_match.group(1))
            self.location = ci_match.group(2)
        
        # Parse timestamps
        timestamp_pattern = r'(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2})'
        timestamps = re.findall(timestamp_pattern, self.log_content)
        if timestamps:
            self.start_time = timestamps[0]
            self.end_time = timestamps[-1]

    def _parse_duration(self, duration_str):
        parts = duration_str.split(':')
        if len(parts) == 3:  # HH:MM:SS
            return int(parts[0]) * 3600 + int(parts[1]) * 60 + float(parts[2])
        elif len(parts) == 2:  # MM:SS
            return int(parts[0]) * 60 + float(parts[1])
        return float(duration_str)

    def calculate_energy_metrics(self):
        total_duration = sum(epoch['duration'] for epoch in self.epochs)
        avg_gpu_power = sum(epoch['gpu_power'] for epoch in self.epochs) / len(self.epochs)
        avg_cpu_power = sum(epoch['cpu_power'] for epoch in self.epochs) / len(self.epochs)
        total_power = avg_gpu_power + avg_cpu_power
        
        # Calculate energy in kWh
        energy_kwh = (total_power * total_duration) / 3600 / 1000
        
        # Calculate CO2 emissions in kg
        co2_kg = (energy_kwh * self.carbon_intensity) / 1000
        
        return {
            'total_duration': total_duration,
            'avg_gpu_power': avg_gpu_power,
            'avg_cpu_power': avg_cpu_power,
            'total_power': total_power,
            'energy_kwh': energy_kwh,
            'co2_kg': co2_kg
        }

    def generate_plots(self):
        # Convert epochs to DataFrame
        df = pd.DataFrame(self.epochs)
        
        # Calculate metrics
        df['cumulative_time'] = df['duration'].cumsum() / 60  # Keep this for x-axis
        df['energy_kwh'] = df['total_power'] * df['duration'] / 3600 / 1000
        df['co2_kg'] = df['energy_kwh'] * self.carbon_intensity / 1000
        
        # Color scheme
        colors = {
            'primary': '#1A237E',
            'accent_blue': '#0072B2',
            'accent_orange': '#E69F00',
            'accent_green': '#009E73',
            'co2': '#D55E00',
            'grid': '#C5CAE9'
        }
        
        # Create figure
        fig = plt.figure(figsize=(10, 6), constrained_layout=True)
        gs = fig.add_gridspec(2, 2, height_ratios=[1.15, 1], width_ratios=[1, 1], hspace=0.1, wspace=0.05)
        
        # Power usage plot (top)
        ax1 = fig.add_subplot(gs[0, :])
        ax1.plot(df['cumulative_time'], df['gpu_power'], label='GPU', linewidth=3.5, marker='o', markersize=5, color=colors['accent_blue'])
        ax1.plot(df['cumulative_time'], df['cpu_power'], label='CPU', linewidth=3.5, marker='s', markersize=5, color=colors['accent_orange'])
        ax1.plot(df['cumulative_time'], df['total_power'], label='Total', linewidth=3.5, linestyle='--', marker='^', markersize=5, color=colors['accent_green'])
        ax1.set_title('Power Usage Over Time', pad=10, fontsize=12, color=colors['primary'])
        ax1.set_xlabel('Time (minutes)', fontsize=10)
        ax1.set_ylabel('Power (W)', fontsize=10)
        ax1.legend(loc='upper right', frameon=True, fancybox=True, shadow=True, fontsize=9)
        ax1.grid(True, alpha=0.5, color=colors['grid'], linestyle='-', linewidth=0.8)
        
        # Energy per epoch plot (bottom left)
        ax2 = fig.add_subplot(gs[1, 0])
        ax2.bar(df['epoch'], df['energy_kwh'], color=colors['accent_blue'], alpha=0.7)
        ax2.set_title('Energy per Epoch', pad=10, fontsize=12, color=colors['primary'])
        ax2.set_xlabel('Epoch', fontsize=10)
        ax2.set_ylabel('Energy (kWh)', fontsize=10)
        ax2.grid(True, alpha=0.5, color=colors['grid'], linestyle='-', linewidth=0.8)
        
        # CO2 emissions per epoch plot (bottom right)
        ax3 = fig.add_subplot(gs[1, 1])
        ax3.bar(df['epoch'], df['co2_kg'], color=colors['co2'], alpha=0.7)
        ax3.set_title('CO2 Emissions per Epoch', pad=10, fontsize=12, color=colors['primary'])
        ax3.set_xlabel('Epoch', fontsize=10)
        ax3.set_ylabel('CO2 (kg)', fontsize=10)
        ax3.grid(True, alpha=0.5, color=colors['grid'], linestyle='-', linewidth=0.8)
        
        # Save plot
        buf = io.BytesIO()
        plt.savefig(buf, format='png', dpi=600, bbox_inches='tight', facecolor='white')
        buf.seek(0)
        plt.close()
        
        return {'combined_plots': buf}

def generate_report_from_log(log_file_path, output_path):
    # Read and parse log
    with open(log_file_path, 'r') as f:
        log_content = f.read()
    
    parser = LogParser(log_content)
    metrics = parser.calculate_energy_metrics()
    plots = parser.generate_plots()
    
    # Create PDF
    doc = SimpleDocTemplate(
        output_path,
        pagesize=letter,
        rightMargin=72,
        leftMargin=72,
        topMargin=24,
        bottomMargin=72
    )
    
    # Styles
    styles = getSampleStyleSheet()
    title_style = ParagraphStyle(
        'CustomTitle',
        parent=styles['Heading1'],
        fontSize=24,
        spaceAfter=10,
        alignment=1,
        textColor=colors.HexColor('#1A237E')
    )
    subtitle_style = ParagraphStyle(
        'CustomSubtitle',
        parent=styles['Normal'],
        fontSize=10,
        spaceAfter=20,
        alignment=1,
        textColor=colors.HexColor('#455A64')
    )
    heading_style = ParagraphStyle(
        'CustomHeading',
        parent=styles['Heading2'],
        fontSize=16,
        spaceAfter=12,
        spaceBefore=12,
        textColor=colors.HexColor('#1A237E')
    )
    normal_style = styles['Normal']
    
    # KPI Table
    kpi_data = [[
        Paragraph(f'<para align="center"><b><font size=16>{metrics["co2_kg"]:.3f} kg</font></b><br/><font size=10>CO2eq Emissions</font><br/><font size=8>Carbon Intensity: {parser.carbon_intensity:.2f} gCO2/kWh</font></para>', normal_style),
        Paragraph(f'<para align="center"><b><font size=16>{metrics["energy_kwh"]:.3f} kWh</font></b><br/><font size=10>Total Energy</font><br/><font size=8>Average Power: {metrics["total_power"]:.1f} W</font></para>', normal_style),
        Paragraph(f'<para align="center"><b><font size=16>{format_duration(metrics["total_duration"])}</font></b><br/><font size=10>Training Duration</font><br/><font size=8>Total Epochs: {len(parser.epochs)}</font></para>', normal_style)
    ]]
    
    kpi_table = Table(kpi_data, colWidths=[2.1*inch, 2.1*inch, 2.1*inch], rowHeights=1.1*inch)
    kpi_table.setStyle(TableStyle([
        ('BACKGROUND', (0, 0), (-1, -1), colors.HexColor('#E8EAF6')),
        ('BOX', (0, 0), (-1, -1), 2, colors.HexColor('#1A237E')),
        ('INNERGRID', (0, 0), (-1, -1), 1, colors.HexColor('#C5CAE9')),
        ('VALIGN', (0, 0), (-1, -1), 'MIDDLE'),
        ('ALIGN', (0, 0), (-1, -1), 'CENTER'),
        ('TOPPADDING', (0, 0), (-1, -1), 10),
        ('BOTTOMPADDING', (0, 0), (-1, -1), 10),
    ]))

    # Location Bar
    location_bar = Table(
        [[Paragraph(f'<b>Location:</b> {parser.location}', normal_style)]],
        colWidths=[6.3*inch]
    )
    location_bar.setStyle(TableStyle([
        ('BACKGROUND', (0, 0), (-1, -1), colors.HexColor('#F5F5FA')),
        ('TEXTCOLOR', (0, 0), (-1, -1), colors.HexColor('#1A237E')),
        ('ALIGN', (0, 0), (-1, -1), 'LEFT'),
        ('FONTNAME', (0, 0), (-1, -1), 'Helvetica'),
        ('FONTSIZE', (0, 0), (-1, -1), 11),
        ('LEFTPADDING', (0, 0), (-1, -1), 10),
        ('RIGHTPADDING', (0, 0), (-1, -1), 10),
        ('TOPPADDING', (0, 0), (-1, -1), 6),
        ('BOTTOMPADDING', (0, 0), (-1, -1), 6),
        ('BOX', (0, 0), (-1, -1), 1, colors.HexColor('#C5CAE9'))
    ]))

    # Parse hardware components
    gpu_name = "-"
    cpu_name = "-"
    if parser.components:
        gpu_match = re.search(r'GPU with device\(s\) ([^\.]+)', parser.components)
        cpu_match = re.search(r'CPU with device\(s\) ([^\.]+)', parser.components)
        if gpu_match:
            gpu_name = gpu_match.group(1).strip()
        if cpu_match:
            cpu_name = cpu_match.group(1).strip()

    # System Power Table
    system_power_data = [
        [Paragraph('<b>System Power</b>', normal_style), ''],
        ['GPU Power', f'{metrics["avg_gpu_power"]:.1f} W ({gpu_name})'],
        ['CPU Power', f'{metrics["avg_cpu_power"]:.1f} W ({cpu_name})'],
        ['Average Power', f'{metrics["total_power"]:.1f} W'],
        ['PUE Coefficient', f'{parser.pue:.2f}']
    ]
    
    system_power_table = Table(system_power_data, colWidths=[2.5*inch, 3.8*inch])
    system_power_table.setStyle(TableStyle([
        ('BACKGROUND', (0, 0), (1, 0), colors.HexColor('#E8EAF6')),
        ('TEXTCOLOR', (0, 0), (0, -1), colors.HexColor('#1A237E')),
        ('ALIGN', (0, 0), (-1, -1), 'LEFT'),
        ('FONTNAME', (0, 0), (-1, 0), 'Helvetica-Bold'),
        ('FONTSIZE', (0, 0), (-1, 0), 12),
        ('BOTTOMPADDING', (0, 0), (-1, 0), 10),
        ('GRID', (0, 0), (-1, -1), 1, colors.HexColor('#C5CAE9')),
        ('SPAN', (0, 0), (1, 0)),
        ('TOPPADDING', (0, 0), (-1, 0), 8),
        ('BOTTOMPADDING', (0, 0), (-1, 0), 8),
    ]))

    # Build document
    story = [
        Paragraph("Carbon Emissions Report", title_style),
        Paragraph(
            f"Generated by CarbonTracker v{parser.version} • {datetime.now().strftime('%Y-%m-%d %H:%M')}",
            subtitle_style
        ),
        Spacer(1, 12),
        Paragraph("Summary", heading_style),
        kpi_table,
        Spacer(1, 12),
        location_bar,
        Spacer(1, 18),
        system_power_table,
        Spacer(1, 20),
        Paragraph("Resource Consumption Analysis", heading_style),
        Image(plots['combined_plots'], width=6*inch, height=3.6*inch)
    ]
    
    doc.build(story)

if __name__ == "__main__":
    import sys
    if len(sys.argv) != 3:
        print("Usage: python report.py <log_file_path> <output_pdf_path>")
        sys.exit(1)
    
    generate_report_from_log(sys.argv[1], sys.argv[2]) 


from carbontracker.components.handler import Handler
from typing import List

class SimulatedCPUHandler(Handler):
    def __init__(self, name: str, tdp: float, utilization: float = 0.5):
        super().__init__(pids=[], devices_by_pid=False)
        if not isinstance(name, str) or not name.strip():
            raise ValueError("CPU name must be a non-empty string.")
        if tdp is None or not isinstance(tdp, (int, float)) or tdp <= 0:
            raise ValueError("CPU TDP must be a positive number.")
        if not isinstance(utilization, (int, float)) or not (0.0 <= utilization <= 1.0):
            raise ValueError("CPU utilization must be between 0.0 and 1.0.")
        self.cpu_brand = name
        self.utilization = utilization
        self.tdp = tdp * utilization

    def devices(self) -> List[str]:
        return [self.cpu_brand]

    def available(self) -> bool:
        return True

    def power_usage(self) -> List[float]:
        return [self.tdp]

    def init(self):
        print(f"Using simulated CPU: {self.cpu_brand} with TDP: {self.tdp:.2f}W (at {self.utilization*100:.0f}% utilization)")

    def shutdown(self):
        pass


import unittest
from unittest.mock import patch
from carbontracker.components.cpu.generic import GenericCPU

class TestGenericCPU(unittest.TestCase):
    def setUp(self):
        self.generic_cpu = GenericCPU(pids=[], devices_by_pid=False)
        self.cpu_power_data = self.generic_cpu.load_cpu_power_data()
        self.average_tdp = self.generic_cpu.calculate_average_tdp()

    @patch('cpuinfo.get_cpu_info')
    def test_known_intel_cpu_tdp(self, mock_cpuinfo):
        known_cpu = "Intel Core i7-9700K"
        known_tdp = 95 / 2  # 50% utilization

        mock_cpuinfo.return_value = {'brand_raw': known_cpu}

        self.generic_cpu = GenericCPU(pids=[], devices_by_pid=False)
        self.generic_cpu.init()

        self.assertEqual(self.generic_cpu.cpu_brand, known_cpu)
        self.assertTrue(self.generic_cpu.available())
        power_usage = self.generic_cpu.power_usage()
        self.assertAlmostEqual(power_usage[0], known_tdp, places=2)

    @patch('cpuinfo.get_cpu_info')
    def test_known_amd_cpu_tdp(self, mock_cpuinfo):
        known_cpu = "6th Gen AMD PRO A10-8770 APU"
        known_tdp = 65 / 2  # 50% utilization

        mock_cpuinfo.return_value = {'brand_raw': known_cpu}

        self.generic_cpu = GenericCPU(pids=[], devices_by_pid=False)
        self.generic_cpu.init()

        self.assertEqual(self.generic_cpu.cpu_brand, known_cpu)
        self.assertTrue(self.generic_cpu.available())
        power_usage = self.generic_cpu.power_usage()
        self.assertAlmostEqual(power_usage[0], known_tdp, places=2)

    @patch('cpuinfo.get_cpu_info')
    def test_unknown_cpu_average_tdp(self, mock_cpuinfo):
        mock_cpuinfo.return_value = {'brand_raw': 'Unknown CPU Model'}

        self.generic_cpu = GenericCPU(pids=[], devices_by_pid=False)
        self.generic_cpu.init()

        self.assertEqual(self.generic_cpu.cpu_brand, 'Unknown CPU Model')
        self.assertTrue(self.generic_cpu.available())
        power_usage = self.generic_cpu.power_usage()
        self.assertAlmostEqual(power_usage[0], self.average_tdp, places=2)  # average_tdp is already halved

    @patch('cpuinfo.get_cpu_info')
    def test_no_cpu_info(self, mock_cpuinfo):
        mock_cpuinfo.return_value = {}

        self.generic_cpu = GenericCPU(pids=[], devices_by_pid=False)
        self.generic_cpu.init()

        self.assertEqual(self.generic_cpu.cpu_brand, 'Unknown CPU')
        self.assertTrue(self.generic_cpu.available())
        power_usage = self.generic_cpu.power_usage()
        self.assertAlmostEqual(power_usage[0], self.average_tdp, places=2)  # average_tdp is already halved

if __name__ == '__main__':
    unittest.main()
